from typing import TYPE_CHECKING, Callable, List, Optional, Text, Union

from applitools.common import AppOutput, Point, Region
from applitools.common.utils import argument_guard, image_utils
from applitools.core import AppOutputWithScreenshot, ServerConnector
from applitools.core.debug import DebugScreenshotsProvider
from applitools.core.extract_text import (
    PATTERN_TEXT_REGIONS,
    BaseOCRRegion,
    ExpectedTextRegion,
    ExtractTextProvider,
    TextRegionSettings,
    TextSettingsData,
)
from applitools.selenium import eyes_selenium_utils
from applitools.selenium.fluent import SeleniumCheckSettings
from applitools.selenium.selenium_eyes import SeleniumEyes

if TYPE_CHECKING:
    from applitools.common.utils.custom_types import (
        AnyWebDriver,
        AnyWebElement,
        CssSelector,
    )
    from applitools.selenium import EyesWebDriver, EyesWebElement


def _element_from_check_settings(driver, check_settings):
    # type: (EyesWebDriver, SeleniumCheckSettings) -> EyesWebElement
    from .webelement import EyesWebElement

    if check_settings.values.target_element:
        element = check_settings.values.target_element
    elif check_settings.values.target_selector:
        by, value = check_settings.values.target_selector
        element = driver.find_element(by, value)
    else:
        raise ValueError
    return EyesWebElement(element, driver)


class OCRRegion(BaseOCRRegion):
    def __init__(self, target):
        # type:(Union[Region,CssSelector,AnyWebElement])->None
        super(OCRRegion, self).__init__(target)
        self.process_app_output = None  # type: Optional[Callable]
        self.app_output_with_screenshot = (
            None
        )  # type: Optional[AppOutputWithScreenshot]

    def add_process_app_output(self, callback):
        # type: (Callable) -> None
        if not callable(callback):
            raise ValueError
        self.process_app_output = callback


class SeleniumExtractTextProvider(ExtractTextProvider):
    def __init__(self, driver, eyes):
        # type: (AnyWebDriver, SeleniumEyes) -> None
        self._driver = driver
        self._eyes = eyes
        self._server_connector = eyes.server_connector  # type: ServerConnector
        self._debug_screenshot_provider = (
            eyes.debug_screenshots_provider
        )  # type: DebugScreenshotsProvider

    def _process_app_output(self, ocr_region):
        settings = SeleniumCheckSettings().fully()
        settings.values.ocr_region = ocr_region

        settings = settings.region(ocr_region.target)

        def app_output_callback(check_settings, region):
            if not check_settings.values.target_region:
                element = _element_from_check_settings(self._driver, check_settings)
                ocr_region.hint = eyes_selenium_utils.get_inner_text(
                    self._driver, element
                )
            app_output = self._eyes._app_output_provider.get_app_output(
                region, self._eyes._last_screenshot, check_settings
            )
            ocr_region.app_output_with_screenshot = app_output

        ocr_region.add_process_app_output(app_output_callback)
        self._eyes.check(settings)

    def get_text(self, *regions):
        # type: (*OCRRegion) -> List[Text]
        result = []
        for ocr_region in regions:
            self._process_app_output(ocr_region)
            image = ocr_region.app_output_with_screenshot.screenshot.image
            screenshot_url = self._server_connector.try_upload_image(
                image_utils.get_bytes(image)
            )
            data = TextSettingsData(
                app_output=AppOutput(
                    screenshot_url=screenshot_url,
                    location=Point.ZERO(),
                    dom_url=ocr_region.app_output_with_screenshot.app_output.dom_url,
                ),
                language=ocr_region._language,
                min_match=ocr_region._min_match,
                regions=[
                    ExpectedTextRegion(
                        0,
                        0,
                        width=image.width,
                        height=image.height,
                        expected=ocr_region._hint,
                    )
                ],
            )
            result.extend(
                self._server_connector.get_text_in_running_session_image(data)
            )
        return result

    def get_text_regions(self, config):
        # type: (TextRegionSettings) -> PATTERN_TEXT_REGIONS
        def get_app_output():
            scale_provider = self._eyes.update_scaling_params()
            viewport_screenshot = self._eyes.get_scaled_cropped_viewport_image(
                scale_provider
            )
            image = image_utils.get_bytes(viewport_screenshot)
            screenshot_url = self._server_connector.try_upload_image(image)

            dom_json = self._eyes._try_capture_dom()
            dom_url = self._eyes._try_post_dom_capture(dom_json)
            return AppOutput(
                dom_url=dom_url, screenshot_url=screenshot_url, location=Point.ZERO()
            )

        argument_guard.not_none(config._patterns)
        data = TextSettingsData(
            app_output=get_app_output(),
            patterns=config._patterns,
            ignore_case=config._ignore_case,
            first_only=config._first_only,
            language=config._language,
        )
        return self._server_connector.get_text_regions_in_running_session_image(data)



import typing as tp

from selenium.common.exceptions import WebDriverException

from applitools.common import logger
from applitools.common.utils.compat import raise_from

if tp.TYPE_CHECKING:
    from selenium.webdriver.remote.webdriver import WebDriver
    from selenium.webdriver.remote.webelement import WebElement

    from applitools.common.utils.custom_types import AnyWebDriver, AnyWebElement
    from applitools.selenium.webdriver import EyesWebDriver


class SwitchToParentIsNotSupported(Exception):
    pass


class EyesWebDriverIsOutOfSync(Exception):
    pass


def ensure_sync_with_underlying_driver(eyes_driver, selenium_driver):
    # type: (EyesWebDriver, WebDriver) -> None
    """
    Checks if frame selected in selenium_driver matches frame_chain in eyes_driver.
    If it doesn't, follows parent frames up to default content and then repeats
    all the chain of frame selections with eyes_driver.
    """
    if not eyes_driver.is_mobile_app:
        try:
            if eyes_driver.frame_chain.size == 0:
                in_sync = _has_no_frame_selected(selenium_driver)
            else:
                selected_frame = eyes_driver.frame_chain.peek
                in_sync = selected_frame.scroll_root_element.is_attached_to_page
        except SwitchToParentIsNotSupported:
            logger.info(
                "Unable to ensure frame chain sync with the underlying driver due to "
                "unsupported 'switch to parent frame' call in the driver"
            )
        if not in_sync:
            try:
                _do_sync_with_underlying_driver(eyes_driver, selenium_driver)
            except Exception as e:
                raise_from(
                    EyesWebDriverIsOutOfSync(
                        "EyesWebDriver frame chain is out of sync. "
                        "Please use web driver returned by Eyes.open call "
                        "for frame switching."
                    ),
                    e,
                )


def _has_no_frame_selected(driver):
    # type: (WebDriver) -> bool
    """
    Predicate that checks if given selenium driver has no frame selected (in other
    words, default content is selected)
    """
    root_element = _current_root_element(driver)
    _swith_to_parent_frame(driver)
    parent_element = _current_root_element(driver)
    if root_element == parent_element:
        return True
    else:
        _switch_to_frame_with_root_element(driver, root_element)
        return False


def _do_sync_with_underlying_driver(eyes_driver, selenium_driver):
    # type: (EyesWebDriver, WebDriverException) -> None
    """
    Goes from currently selected frame in selenium_driver up to
    the topmost default content and remembers the trace.
    Then follows that trace with eyes_driver to get to original frame and have
    frame_chain synchronized.
    """
    root_elements_trace = [_current_root_element(selenium_driver)]
    while True:
        _swith_to_parent_frame(selenium_driver)
        root_element = _current_root_element(selenium_driver)
        if root_element != root_elements_trace[-1]:
            root_elements_trace.append(root_element)
        else:
            root_elements_trace.pop()
            break
    eyes_driver.frame_chain.clear()
    for root_element in reversed(root_elements_trace):
        _switch_to_frame_with_root_element(eyes_driver, root_element)


def _switch_to_frame_with_root_element(driver, root_element):
    # type: (AnyWebDriver, AnyWebElement) -> None
    """
    Tries to find and switch given driver to a frame that has given root_element
    as it's root element.
    """
    for frame in driver.find_elements_by_xpath("//iframe|//frame"):
        driver.switch_to.frame(frame)
        if _current_root_element(driver) == root_element:
            break
        else:
            _swith_to_parent_frame(driver)
    else:
        raise RuntimeError("Failed to reach element {}".format(root_element))


def _current_root_element(driver):
    # type: (WebDriver) -> WebElement
    """
    Finds root WebElement within currently selected frame in given driver.
    Usually it's the 'html' element.
    """
    return driver.find_element_by_xpath("/*")


def _swith_to_parent_frame(driver):
    # type: (WebDriver) -> None
    """
    Switches given driver to the parent frame or raises specific exception
    """
    try:
        driver.switch_to.parent_frame()
    except WebDriverException as e:
        if "Method is not implemented" in e.msg:
            raise_from(SwitchToParentIsNotSupported, e)
        else:
            raise



import concurrent
import itertools
import sys
import threading
import typing
from collections import Counter, deque
from concurrent.futures import ThreadPoolExecutor
from enum import Enum
from time import time

import attr

from applitools.common import TestResults, TestResultsSummary, logger
from applitools.common.client_event import ClientEvent, TraceLevel
from applitools.common.utils import datetime_utils, json_utils
from applitools.common.utils.compat import Queue
from applitools.core import EyesRunner
from applitools.selenium.visual_grid.rendreing_service import RenderingService
from applitools.selenium.visual_grid.running_test import COMPLETED

from .helpers import collect_test_results, wait_till_tests_completed
from .resource_cache import PutCache, ResourceCache

if typing.TYPE_CHECKING:
    from typing import Dict, List, Text, Union

    from applitools.core import ServerConnector
    from applitools.selenium.visual_grid import RunningTest, VGTask, VisualGridEyes


class RunnerOptions(object):
    def __init__(self):
        self._test_concurrency = _TestConcurrency()

    def test_concurrency(self, value):
        # type: (int) -> RunnerOptions
        self._test_concurrency = _TestConcurrency(
            _TestConcurrency.Kind.TEST_CONCURRENCY, value
        )
        return self

    def get_test_concurrency(self):
        # type: () -> _TestConcurrency
        return self._test_concurrency


@attr.s
class _TestConcurrency(object):
    class Kind(Enum):
        DEFAULT = "defaultConcurrency"
        TEST_CONCURRENCY = "testConcurrency"
        LEGACY = "concurrency"

    LEGACY_FACTOR = 5
    kind = attr.ib(default=Kind.DEFAULT)  # type: Text
    value = attr.ib(default=5)  # type: int


class _ResourceCollectionService(object):
    def __init__(self):
        self._queue = Queue()
        self._thread = threading.Thread(target=self._resource_collection, args=())
        self._thread.daemon = True
        self._thread.setName(self.__class__.__name__)
        self._thread.start()

    def _resource_collection(self):
        while True:
            task = self._queue.get()
            if task is None:
                break
            task()

    def add_task(self, task):
        self._queue.put(task)

    def shutdown(self):
        self._queue.put(None)


class VisualGridRunner(EyesRunner):
    def __init__(self, options_or_concurrency=RunnerOptions()):
        # type: (Union[RunnerOptions, int]) -> None
        if isinstance(options_or_concurrency, int):
            self._concurrency = _TestConcurrency(
                _TestConcurrency.Kind.LEGACY,
                options_or_concurrency * _TestConcurrency.LEGACY_FACTOR,
            )
        else:
            self._concurrency = options_or_concurrency.get_test_concurrency()

        super(VisualGridRunner, self).__init__()
        self._runner_started_log_sent = False
        self._last_states_logging_time = time()
        self._all_test_results = {}  # type: Dict[RunningTest, TestResults]

        kwargs = {}
        if sys.version_info[:2] >= (3, 6):
            kwargs["thread_name_prefix"] = "VGR-Executor"

        self.resource_cache = ResourceCache()  # type:ResourceCache
        self.put_cache = PutCache()  # type: PutCache
        self.all_eyes = []  # type: List[VisualGridEyes]
        self.still_running = True  # type: bool

        self._executor = ThreadPoolExecutor(
            max_workers=self._concurrency.value, **kwargs
        )
        self._future_to_task = ResourceCache()  # type:ResourceCache
        self._parallel_tests = []  # type: List[RunningTest]
        thread = threading.Thread(target=self._run, args=())
        thread.setName(self.__class__.__name__)
        thread.daemon = True
        thread.start()
        self._thread = thread

        self._resource_collection_service = _ResourceCollectionService()
        self.rendering_service = RenderingService()

    def add_resource_collection_task(self, task):
        self._resource_collection_service.add_task(task)

    def __del__(self):
        self._stop()

    def aggregate_result(self, test, test_result):
        # type: (RunningTest, TestResults) -> None
        self.logger.debug(
            "aggregate_result called", test_id=test.test_uuid, test_result=test_result
        )
        self._all_test_results[test] = test_result

    def open(self, eyes):
        # type: (VisualGridEyes) -> None
        self.all_eyes.append(eyes)
        self.logger.debug("VisualGridRunner.open()", eyes_id=id(eyes))
        if not self._runner_started_log_sent:
            self._runner_started_log_sent = True
            self._send_runner_started_log_message(eyes.server_connector)

    def _current_parallel_tests(self):
        concurrent_sessions = self._concurrency.value
        parallel_tests = [
            test for test in self._parallel_tests if test.state != COMPLETED
        ]
        tests_left = len(parallel_tests)
        if tests_left < concurrent_sessions:
            tests_to_add = concurrent_sessions - tests_left
            parallel_tests += self._get_n_not_completed_tests(tests_to_add)
            self._parallel_tests = parallel_tests
        return self._parallel_tests

    def _run(self):
        self.logger.debug("VisualGridRunner.run()")
        for test_queue in self._get_parallel_tests_by_round_robin():
            try:
                task = test_queue.popleft()
                self.logger.debug("VisualGridRunner got task", task=task)
            except IndexError:
                datetime_utils.sleep(10, msg="Waiting for task", verbose=False)
                continue
            future = self._executor.submit(task)
            self._future_to_task[future] = task
        self.logger.debug("VisualGridRunner.run() done")

    def _stop(self):
        # type: () -> None
        while any(r.state != COMPLETED for r in self._get_all_running_tests()):
            datetime_utils.sleep(500, msg="Waiting for finishing tests in stop")
        self.still_running = False
        for future in concurrent.futures.as_completed(self._future_to_task):
            task = self._future_to_task[future]
            try:
                future.result()
            except Exception:
                self.logger.exception("Task generated an exception", task=task)

        self.put_cache.shutdown()
        self._resource_collection_service.shutdown()
        self.resource_cache.executor.shutdown()
        self._executor.shutdown()
        self._thread.join()
        self.rendering_service.shutdown()

    def _get_all_test_results_impl(self, should_raise_exception=True):
        # type: (bool) -> TestResultsSummary
        wait_till_tests_completed(self._get_all_running_tests)

        # finish processing of all tasks and shutdown threads
        self._stop()

        all_results = collect_test_results(
            self._all_test_results, should_raise_exception
        )
        return TestResultsSummary(all_results)

    def _get_all_running_tests(self):
        # type: ()-> List[RunningTest]
        tests = list(itertools.chain.from_iterable(e.test_list for e in self.all_eyes))
        if time() - self._last_states_logging_time > 15:
            self._last_states_logging_time = time()
            # print states every 15 seconds
            counter = Counter(t.state for t in tests)
            self.logger.info("Current tests states", **counter)
        return tests

    def _get_n_not_completed_tests(self, n):
        all_tests = self._get_all_running_tests()
        not_opened = [
            test
            for test in all_tests
            if test.state != COMPLETED and test not in self._parallel_tests
        ]
        return not_opened[:n]

    def _get_parallel_tests_by_round_robin(self):
        # type: () -> List[VGTask]
        done = False
        next_test = 0
        while not done:
            current_tests = self._current_parallel_tests()
            if current_tests:
                index = next_test % len(current_tests)
                yield current_tests[index].queue
                next_test += 1
            else:
                if self.still_running:
                    yield deque()
                else:
                    done = True

    def _send_runner_started_log_message(self, server_connector):
        # type: (ServerConnector) -> None
        message = {
            "type": "runnerStarted",
            self._concurrency.kind.value: self._concurrency.value,
            # ... other properties like node version, os, architecture, etc.
        }
        server_connector.send_logs(
            ClientEvent(TraceLevel.Notice, json_utils.to_json(message))
        )



from __future__ import unicode_literals

import json
import typing
from abc import abstractmethod
from enum import Enum
from time import time

import attr
from selenium.webdriver.common.by import By

from applitools.common.utils import ABC, datetime_utils
from applitools.common.utils.json_utils import JsonInclude, to_json
from applitools.selenium import resource
from applitools.selenium.fluent import FrameLocator

if typing.TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Text, Union

    from structlog import BoundLogger

    from applitools.selenium import EyesWebDriver


MAX_CHUNK_BYTES_IOS = 10 * 1024 * 1024
MAX_CHUNK_BYTES_GENERIC = 50 * 1024 * 1024


class DomSnapshotFailure(Exception):
    pass


class DomSnapshotScriptError(DomSnapshotFailure):
    pass


class DomSnapshotTimeout(DomSnapshotFailure):
    pass


def create_dom_snapshot(
    driver,  # type: EyesWebDriver
    logger,  # type: BoundLogger
    dont_fetch_resources,  # type: bool
    skip_resources,  # type: Optional[List[Text]]
    timeout_ms,  # type: int
    cross_origin_rendering,  # type: bool
    use_cookies,  # type: bool
):
    # type: (...) -> Dict
    is_ie = driver.user_agent.is_internet_explorer
    script_type = DomSnapshotScriptForIE if is_ie else DomSnapshotScriptGeneric
    script = script_type(driver)
    is_ios = "ios" in driver.desired_capabilities.get("platformName", "").lower()
    chunk_byte_length = MAX_CHUNK_BYTES_IOS if is_ios else MAX_CHUNK_BYTES_GENERIC
    snapshotter = RecursiveSnapshotter(
        driver,
        script,
        logger,
        timeout_ms,
        chunk_byte_length,
        cross_origin_rendering,
        use_cookies,
        dont_fetch_resources=dont_fetch_resources,
        skip_resources=skip_resources,
        serialize_resources=True,
    )
    try:
        return snapshotter.create_cross_frames_dom_snapshots()
    except Exception:
        logger.info(
            "Failed to create dom-snapshot, retrying ignoring failing frames",
            exc_info=True,
        )
        snapshotter.should_skip_failed_frames = True
        return snapshotter.create_cross_frames_dom_snapshots()


@attr.s
class ProcessPageArgs(object):
    show_logs = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    use_session_cache = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    dont_fetch_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    fetch_timeout = attr.ib(
        type=int, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[int]
    skip_resources = attr.ib(
        type=list, metadata={JsonInclude.NON_NONE: True}
    )  # type: List[Text]
    compress_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]
    serialize_resources = attr.ib(
        type=bool, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[bool]


@attr.s
class PollResultArgs(object):
    chunk_byte_length = attr.ib(
        type=int, metadata={JsonInclude.NON_NONE: True}
    )  # type: Optional[int]


class ProcessPageStatus(Enum):
    SUCCESS = "SUCCESS"
    WIP = "WIP"
    ERROR = "ERROR"
    SUCCESS_CHUNKED = "SUCCESS_CHUNKED"


@attr.s
class ProcessPageResult(object):
    status = attr.ib(type=ProcessPageStatus)  # type: ProcessPageStatus
    done = attr.ib(type=bool, default=None)  # type: Optional[bool]
    error = attr.ib(type=str, default=None)  # type: Optional[Text]
    value = attr.ib(default=None)  # type: Optional[Union[Dict,Text]]

    @classmethod
    def from_json(cls, status_json):
        # attr_from_json doesn't support untyped dict fields for now
        status_json = json.loads(status_json)
        return cls(
            status=ProcessPageStatus(status_json["status"]),
            done=status_json.get("done"),
            error=status_json.get("error"),
            value=status_json.get("value"),
        )


class DomSnapshotScript(ABC):
    """Base class for different flavors of dom-snapshot script"""

    @abstractmethod
    def process_page_script_code(self, args):
        # type: (Text) -> Text
        pass

    @abstractmethod
    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        pass

    def __init__(self, driver):
        # type: (EyesWebDriver) -> None
        self._driver = driver

    def run(
        self,
        show_logs=None,  # type: Optional[bool]
        use_session_cache=None,  # type: Optional[bool]
        dont_fetch_resources=None,  # type: Optional[bool]
        fetch_timeout=None,  # type: Optional[int]
        skip_resources=None,  # type: Optional[List[Text]]
        compress_resources=None,  # type: Optional[bool]
        serialize_resources=None,  # type: Optional[bool]
    ):
        # type: (...) -> ProcessPageResult
        return self._run_script(
            locals(), ProcessPageArgs, self.process_page_script_code
        )

    def poll_result(self, chunk_byte_length=None):
        # type: (Optional[int]) -> ProcessPageResult
        return self._run_script(locals(), PollResultArgs, self.poll_result_script_code)

    def _run_script(self, args, args_type, code_gen_func):
        # type: (Dict[Text, Any], type, Callable[[Text], Text]) -> ProcessPageResult
        args = {k: v for k, v in args.items() if k != "self"}
        code = code_gen_func(to_json(args_type(**args)))
        result_json = self._driver.execute_script(code)
        return ProcessPageResult.from_json(result_json)


class DomSnapshotScriptGeneric(DomSnapshotScript):
    _script_code = resource.get_resource("processPagePoll.js")
    _poll_result_code = resource.get_resource("pollResult.js")

    def process_page_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __processPagePoll({});".format(self._script_code, args)

    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __pollResult({});".format(self._poll_result_code, args)


class DomSnapshotScriptForIE(DomSnapshotScript):
    _script_code = resource.get_resource("processPagePollForIE.js")
    _poll_result_code = resource.get_resource("pollResultForIE.js")

    def process_page_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __processPagePollForIE({});".format(self._script_code, args)

    def poll_result_script_code(self, args):
        # type: (Text) -> Text
        return "{} return __pollResultForIE({});".format(self._poll_result_code, args)


class RecursiveSnapshotter(object):
    POLL_INTERVAL_MS = 1000

    def __init__(
        self,
        driver,  # type: EyesWebDriver
        script,  # type: DomSnapshotScript
        logger,  # type: BoundLogger
        timeout_ms,  # type: int
        chunk_byte_length,  # type: int
        cross_origin_rendering,  # type: bool
        use_cookies,  # type: bool
        **script_args  # type: Any
    ):
        self.should_skip_failed_frames = False
        self._driver = driver
        self._script = script
        self._logger = logger
        self._deadline_time = time() + datetime_utils.to_sec(timeout_ms)
        self._chunk_byte_length = chunk_byte_length
        self._cross_origin_rendering = cross_origin_rendering
        self._use_cookies = use_cookies
        self._script_args = script_args

    def create_cross_frames_dom_snapshots(self):
        dom = self._create_dom_snapshot_loop()
        self._process_dom_snapshot_frames(dom)
        return dom

    def _create_dom_snapshot_loop(self):
        # type: () -> Dict
        chunks = []
        result = self._script.run(**self._script_args)
        while result.status is ProcessPageStatus.WIP or (
            result.status is ProcessPageStatus.SUCCESS_CHUNKED and not result.done
        ):
            if time() > self._deadline_time:
                raise DomSnapshotTimeout
            result = self._script.poll_result(self._chunk_byte_length)
            if result.status is ProcessPageStatus.WIP:
                datetime_utils.sleep(
                    self.POLL_INTERVAL_MS,
                    "Waiting for the end of DOM extraction",
                )
            elif result.status is ProcessPageStatus.SUCCESS_CHUNKED:
                self._logger.info(
                    "Snapshot chunk {}, {}B".format(len(chunks), len(result.value))
                )
                chunks.append(result.value)
        if result.status is ProcessPageStatus.SUCCESS:
            return result.value
        elif result.status.SUCCESS_CHUNKED and result.done:
            return json.loads("".join(chunks))
        elif result.status is ProcessPageStatus.ERROR:
            raise DomSnapshotScriptError(result.error)
        else:
            raise DomSnapshotFailure("Unexpected script result", result)

    def _process_dom_snapshot_frames(self, dom):
        # type: (Dict) -> None
        if self._use_cookies:
            dom["cookies"] = self._driver.get_cookies()
        for frame in dom["frames"]:
            selector = frame.get("selector", None)
            if not selector:
                self._logger.warning("inner frame with null selector")
                continue
            try:
                with self._driver.switch_to.frame_and_back(
                    FrameLocator(frame_selector=[By.CSS_SELECTOR, selector])
                ):
                    self._process_dom_snapshot_frames(frame)
            except Exception:
                if self.should_skip_failed_frames:
                    self._logger.warning(
                        "failed switching to frame",
                        frame_selector=selector,
                        exc_info=True,
                    )
                else:
                    raise
        if self._cross_origin_rendering:
            self._snapshot_and_process_cors_frames(dom)

    def _snapshot_and_process_cors_frames(self, dom):
        # type: (Dict) -> None
        for frame in dom["crossFrames"]:
            selector = frame.get("selector", None)
            if not selector:
                self._logger.warning("cross frame with null selector")
                continue
            frame_index = frame["index"]
            try:
                with self._driver.switch_to.frame_and_back(
                    FrameLocator(frame_selector=[By.CSS_SELECTOR, selector])
                ):
                    frame_dom = self.create_cross_frames_dom_snapshots()
                    dom.setdefault("frames", []).append(frame_dom)
                    frame_url = frame_dom["url"]
                    dom["cdt"][frame_index]["attributes"].append(
                        {"name": "data-applitools-src", "value": frame_url}
                    )
                    self._logger.info(
                        "Created cross origin frame snapshot {}".format(frame_url)
                    )
                    self._process_dom_snapshot_frames(frame_dom)
            except Exception:
                if self.should_skip_failed_frames:
                    self._logger.warning(
                        "failed extracting and processing cross frame",
                        frame_selector=selector,
                        exc_info=True,
                    )
                else:
                    raise


def has_cross_subframes(dom):
    if dom["crossFrames"]:
        return True
    return any(has_cross_subframes(frame) for frame in dom["frames"])



import typing
from copy import deepcopy
from threading import Condition, Thread
from time import time

import attr

from applitools.common import EyesError, RenderStatus, logger
from applitools.common.utils import datetime_utils

if typing.TYPE_CHECKING:
    from typing import Callable, List, Text

    from applitools.common import RenderRequest, RenderStatusResults
    from applitools.core import ServerConnector


MAX_STATUS_CHECK_SECONDS = 60 * 60


class RenderingService(object):
    def __init__(self):
        self._server_connector = None  # type: ServerConnector
        self._shutdown = False
        self._render_tasks = []  # type: List[_Render]
        self._have_render_tasks = Condition()
        self._status_tasks = []  # type: List[_Status]
        self._have_status_tasks = Condition()
        self._render_request_thread = Thread(target=self._run_render_requests)
        self._render_request_thread.daemon = True
        self._render_status_thread = Thread(target=self._run_render_status_requests)
        self._render_status_thread.daemon = True
        self._render_request_thread.start()
        self._render_status_thread.start()

    def maybe_set_server_connector(self, server_connector):
        # type: (ServerConnector) -> None
        if not self._server_connector:
            self._server_connector = server_connector

    def render(self, render_request, on_success, on_error):
        # type: (RenderRequest, Callable, Callable) -> None
        with self._have_render_tasks:
            self._render_tasks.append(_Render(render_request, on_success, on_error))
            self._have_render_tasks.notify()

    def shutdown(self):
        self._shutdown = True
        with self._have_render_tasks:
            self._have_render_tasks.notify()
        with self._have_status_tasks:
            self._have_status_tasks.notify()
        self._render_request_thread.join()
        self._render_status_thread.join()

    def _run_render_requests(self):
        while True:
            with self._have_render_tasks:
                while not (self._shutdown or self._render_tasks):
                    self._have_render_tasks.wait()
                if self._shutdown:
                    break
                render_tasks, self._render_tasks = self._render_tasks, []
            requests = [r.request for r in render_tasks]
            logger.debug("render call with requests: {}".format(requests))
            try:
                responses = self._server_connector.render(*requests)
            except Exception as e:
                for task in render_tasks:
                    task.on_error(e)
                continue
            statuses = []
            for task, r in zip(render_tasks, responses):
                if (
                    r.render_status == RenderStatus.NEED_MORE_RESOURCE
                    or r.need_more_dom
                ):
                    task.on_error(EyesError("Some resources aren't uploaded"))
                else:
                    statuses.append(
                        _Status(r.render_id, task.on_success, task.on_error)
                    )
            with self._have_status_tasks:
                self._status_tasks.extend(statuses)
                self._have_status_tasks.notify()

    def _run_render_status_requests(self):
        status_tasks = []
        while True:
            with self._have_status_tasks:
                while not (self._shutdown or status_tasks or self._status_tasks):
                    self._have_status_tasks.wait()
                if self._shutdown:
                    break
                new_tasks, self._status_tasks = self._status_tasks, []
            status_tasks.extend(new_tasks)
            ids = [t.render_id for t in status_tasks]
            logger.debug("render_status_by_id call with ids: {}".format(ids))
            try:

                statuses = self._server_connector.render_status_by_id(*ids)
            except Exception as e:
                for task in status_tasks:
                    task.on_error(e)
                continue
            rendering = []
            current_time = time()
            for status, task in zip(statuses, status_tasks):
                if status.status == RenderStatus.RENDERING:
                    if current_time > task.timeout_time:
                        task.on_error(EyesError("Render time out"))
                    else:
                        rendering.append(task)
                else:
                    task.on_success(status)
            status_tasks = rendering
            if status_tasks:
                datetime_utils.sleep(1000, "Render status polling delay")


@attr.s
class _Render(object):
    request = attr.ib()  # type: RenderRequest
    on_success = attr.ib()  # type: Callable[[RenderStatusResults], None]
    on_error = attr.ib()  # type: Callable[[Exception], None]


@attr.s
class _Status(object):
    render_id = attr.ib()  # type: Text
    on_success = attr.ib()  # type: Callable[[RenderStatusResults], None]
    on_error = attr.ib()  # type: Callable[[Exception], None]
    timeout_time = attr.ib(
        factory=lambda: time() + MAX_STATUS_CHECK_SECONDS
    )  # type: int



import time
import typing
from collections import deque
from itertools import chain

import attr

from applitools.common import (
    EyesError,
    RenderInfo,
    RenderRequest,
    RGridDom,
    VGResource,
    VisualGridSelector,
    logger,
)
from applitools.common.utils import apply_base_url
from applitools.common.utils.compat import urlparse
from applitools.selenium.parsers import collect_urls_from_

from .vg_task import VGTask

if typing.TYPE_CHECKING:
    from typing import (
        Any,
        Callable,
        Dict,
        Iterable,
        List,
        NoReturn,
        Optional,
        Text,
        Tuple,
    )

    from applitools.common import Region, RenderingInfo
    from applitools.core import ServerConnector
    from applitools.selenium.visual_grid import PutCache, ResourceCache, RunningTest


@attr.s(hash=True)
class ResourceCollectionTask(VGTask):
    MAX_FAILS_COUNT = 5
    MAX_ITERATIONS = 2400  # poll_render_status for 1 hour

    script = attr.ib(hash=False, repr=False)  # type: Dict[str, Any]
    resource_cache = attr.ib(hash=False, repr=False)  # type: ResourceCache
    put_cache = attr.ib(hash=False, repr=False)  # type: PutCache
    server_connector = attr.ib(hash=False, repr=False)  # type: ServerConnector
    rendering_info = attr.ib()  # type: RenderingInfo
    region_selectors = attr.ib(
        hash=False, factory=list
    )  # type: List[List[VisualGridSelector]]
    size_mode = attr.ib(default=None)
    region_to_check = attr.ib(hash=False, default=None)  # type: Region
    script_hooks = attr.ib(hash=False, default=None)  # type: Optional[Dict]
    agent_id = attr.ib(default=None)  # type: Optional[Text]
    selector = attr.ib(hash=False, default=None)  # type: Optional[VisualGridSelector]
    func_to_run = attr.ib(default=None, hash=False, repr=False)  # type: Callable
    running_tests = attr.ib(hash=False, factory=list)  # type: List[RunningTest]
    request_options = attr.ib(hash=False, factory=dict)  # type: Dict[str, Any]

    def __attrs_post_init__(self):
        # type: () -> None
        self.full_request_resources = {}  # type: Dict[Text, VGResource]
        self.func_to_run = lambda: self.prepare_data_for_rg(
            self.script
        )  # type: Callable

    def prepare_data_for_rg(self, data):
        # type: (Dict) -> Dict[RunningTest,RenderRequest]
        dom = self.parse_frame_dom_resources(data)
        render_requests = self.prepare_rg_requests(dom, self.full_request_resources)
        logger.debug(
            "exit - returning render_request array of length: {}".format(
                len(render_requests)
            )
        )
        render_request = list(render_requests.values())[0]
        logger.debug("Uploading missing resources")
        self.put_cache.put(
            list(render_request.resources.values()) + [render_request.dom.resource],
            self.server_connector,
        )
        return render_requests

    def prepare_rg_requests(self, dom, request_resources):
        # type: (RGridDom, Dict) -> Dict[RunningTest,RenderRequest]
        if self.size_mode == "region" and self.region_to_check is None:
            raise EyesError("Region to check should be present")
        if self.size_mode == "selector" and not isinstance(
            self.selector, VisualGridSelector
        ):
            raise EyesError("Selector should be present")
        requests = {}
        region = None
        for running_test in self.running_tests:
            if self.region_to_check:
                region = dict(
                    x=self.region_to_check.x,
                    y=self.region_to_check.y,
                    width=self.region_to_check.width,
                    height=self.region_to_check.height,
                )
            r_info = RenderInfo.from_(
                size_mode=self.size_mode,
                selector=self.selector,
                region=region,
                render_browser_info=running_test.browser_info,
            )

            requests[running_test] = RenderRequest(
                webhook=self.rendering_info.results_url,
                agent_id=self.agent_id,
                url=dom.url,
                stitching_service=self.rendering_info.stitching_service_url,
                dom=dom,
                resources=request_resources,
                render_info=r_info,
                renderer=running_test.eyes.renderer,
                browser_name=running_test.browser_info.browser,
                platform_name=running_test.browser_info.platform,
                script_hooks=self.script_hooks,
                selectors_to_find_regions_for=list(chain(*self.region_selectors)),
                send_dom=running_test.configuration.send_dom,
                options=self.request_options,
            )
        return requests

    def parse_frame_dom_resources(self, data):  # noqa
        # type: (Dict) -> RGridDom
        base_url = data["url"]
        resource_urls = data.get("resourceUrls", [])
        all_blobs = data.get("blobs", [])
        frames = data.get("frames", [])
        cookies = data.get("cookies", [])
        logger.debug(
            "parse_frame_dom_resources() call",
            base_url=base_url,
            blobs_count=len(all_blobs),
            resource_urls_count=len(resource_urls),
            frames_count=len(frames),
            cookies_count=len(cookies),
        )

        def find_child_resource_urls(content_type, content, resource_url):
            # type: (Optional[Text], bytes, Text) -> NoReturn
            logger.debug(
                "find_child_resource_urls({0}, {1}) call".format(
                    content_type, resource_url
                )
            )
            if not content_type:
                logger.debug("content_type is empty. Skip handling of resources")
                return []
            return [
                apply_base_url(url, base_url, resource_url)
                for url in collect_urls_from_(content_type, content)
            ]

        frame_request_resources = {}
        for f_data in frames:
            f_data["url"] = apply_base_url(f_data["url"], base_url)
            frame_request_resources[f_data["url"]] = self.parse_frame_dom_resources(
                f_data
            ).resource

        urls_to_fetch = set(resource_urls)
        for blob in all_blobs:
            resource = VGResource.from_blob(blob, find_child_resource_urls)
            if resource.url.rstrip("#") == base_url:
                continue
            frame_request_resources[resource.url] = resource
            urls_to_fetch |= set(resource.child_resource_urls)

        resources_and_their_children = fetch_resources_recursively(
            urls_to_fetch,
            cookies,
            self.server_connector,
            self.resource_cache,
            find_child_resource_urls,
        )
        frame_request_resources.update(resources_and_their_children)
        self.full_request_resources.update(frame_request_resources)
        return RGridDom(
            url=base_url, dom_nodes=data["cdt"], resources=frame_request_resources
        )


def fetch_resources_recursively(
    urls,  # type: Iterable[Text]
    cookies,  # type: Dict
    eyes_connector,  # type: ServerConnector
    resource_cache,  # type: ResourceCache
    find_child_resource_urls,  # type: Callable[[Text, bytes, Text],List[Text]]
):
    # type: (...) -> Iterable[Tuple[Text, VGResource]]
    def get_resource(link):
        logger.debug("get_resource({0}) call".format(link))
        matching_cookies = [c for c in cookies if is_cookie_for_url(c, link)]
        response = eyes_connector.download_resource(link, matching_cookies)
        return VGResource.from_response(link, response, find_child_resource_urls)

    def schedule_fetch(urls):
        for url in urls:
            if url not in seen_urls:
                seen_urls.add(url)
                downloading = resource_cache.fetch_and_store(url, get_resource)
                if downloading:  # going to take time, add to the queue end
                    fetched_urls_deque.appendleft(url)
                else:  # resource is already in cache, add to the queue front
                    fetched_urls_deque.append(url)

    fetched_urls_deque = deque()
    seen_urls = set()
    schedule_fetch(urls)
    while fetched_urls_deque:
        url = fetched_urls_deque.pop()
        resource = resource_cache[url]
        if resource is None:
            logger.debug("No response for {}".format(url))
        else:
            schedule_fetch(resource.child_resource_urls)
            yield url, resource


def is_cookie_for_url(cookie, url):
    # type: (Dict, Text) -> bool
    url = urlparse(url)
    if cookie["secure"] and url.scheme != "https":
        return False
    if "expiry" in cookie and time.time() > cookie["expiry"]:
        return False
    subdomains_allowed = cookie["domain"][0] == "."
    domain = cookie["domain"][1:] if subdomains_allowed else cookie["domain"]
    domain_match = url.hostname == domain
    subdomain_match = url.hostname.endswith("." + domain)
    if not (domain_match or subdomains_allowed and subdomain_match):
        return False
    path = cookie["path"]
    path = path[:-1] if path[-1] == "/" else path
    if url.path != path and not url.path.startswith(path + "/"):
        return False
    return True



from __future__ import division

import typing
from math import ceil

import attr

from applitools.common import Point
from applitools.common.utils.image_utils import image_from_bytes

if typing.TYPE_CHECKING:
    from typing import List, Optional, Tuple

    from PIL.Image import Image

    from applitools.common.utils.custom_types import AnyWebDriver


def device_viewport_location(driver):
    # type: (AnyWebDriver) -> Point
    pattern = add_page_marker(driver)
    try:
        image = image_from_bytes(driver.get_screenshot_as_png())
    finally:
        remove_page_marker(driver)
    return find_pattern(image, pattern)


@attr.s
class Pattern(object):
    offset = attr.ib()  # type: int
    size = attr.ib()  # type: int
    mask = attr.ib()  # type: List[int]


def add_page_marker(driver):
    # type: (AnyWebDriver, float) -> Pattern
    pattern_dict = driver.execute_script(_ADD_PAGE_MARKER_JS)
    return Pattern(**pattern_dict)


def remove_page_marker(driver):
    # type: (AnyWebDriver) -> None
    driver.execute_script(_CLEANUP_PAGE_MARKER_JS)


def find_pattern(image, pattern):
    # type: (Image, Pattern) -> Optional[Point]
    image = image.convert("RGB")
    for y in range(image.height // 2):  # Look for pattern in top left quadrant only
        for x in range(image.width // 2):
            if _is_pattern(image, x, y, pattern):
                return Point(x - pattern.offset, y - pattern.offset)


def _is_pattern(image, x, y, pattern):
    # type: (Image, int, int, Pattern) -> bool
    chunk_midsize = int(ceil(pattern.size / 2))
    # Try to check if centers of all chunks have matching color
    # makes significant performance improvement on dark images
    for chunk_index, chunk_color in enumerate(pattern.mask):
        chunk_center = x + chunk_midsize, y + chunk_midsize + chunk_index * pattern.size
        if _pixel_color_at(image, chunk_center, 15) != chunk_color:
            return False
    for chunk_index, chunk_color in enumerate(pattern.mask):
        threshold = 40
        # Check if all pixels of the chunk have matching color
        # Check is done from borders to center, reducing threshold closer to center
        for round in range(chunk_midsize):
            round_x = x + round
            round_y = y + round + chunk_index * pattern.size
            side_length = pattern.size - round * 2
            threshold = max(15, threshold - 10)
            for i in range(side_length - 1):
                top = (round_x + i, round_y)
                left = (round_x, round_y + i + 1)
                bottom = (round_x + i + 1, round_y + side_length - 1)
                right = (round_x + side_length - 1, round_y + i)
                for pixel in top, right, left, bottom:
                    if _pixel_color_at(image, pixel, threshold) != chunk_color:
                        return False
    return True


def _pixel_color_at(image, xy, threshold):
    # type: (Image, Tuple[int, int], int) -> int
    r, g, b = image.getpixel(xy)
    white = 255 - threshold
    if r >= white and g >= white and b >= white:
        return 1
    elif r <= threshold and g <= threshold and b <= threshold:
        return 0
    else:
        return -1


_SET_ELEMENT_STYLE_PROPERTIES_JS = """
function __setElementStyleProperties([element, properties] = []) {
  const keys = Object.keys(properties).sort()
  const original = keys.reduce((original, prop) => {
    original[prop] = {
      value: element.style.getPropertyValue(prop),
      important: Boolean(element.style.getPropertyPriority(prop)),
    }
    return original
  }, {})

  keys.forEach(prop => {
    element.style.setProperty(
      prop,
      typeof properties[prop] === 'string' || !properties[prop]
        ? properties[prop]
        : properties[prop].value,
      properties[prop] && properties[prop].important ? 'important' : '',
    )
  })
  return original
}
"""

_ADD_PAGE_MARKER_JS = "{} {} return __addPageMarker()".format(
    _SET_ELEMENT_STYLE_PROPERTIES_JS,
    """
function __addPageMarker() {
  const marker = document.createElement('div')
  const contrast = document.createElement('div')
  marker.appendChild(contrast)
  document.body.appendChild(marker)
  marker.setAttribute('data-applitools-marker', '')

  marker.style.setProperty('position', 'fixed', 'important')
  marker.style.setProperty('top', '0', 'important')
  marker.style.setProperty('left', '0', 'important')
  marker.style.setProperty('width', '3px', 'important')
  marker.style.setProperty('height', '9px', 'important')
  marker.style.setProperty('box-sizing', 'content-box', 'important')
  marker.style.setProperty('border', '1px solid rgb(127,127,127)', 'important')
  marker.style.setProperty('background', 'rgb(0,0,0)', 'important')
  marker.style.setProperty('z-index', '999999999', 'important')

  contrast.style.setProperty('width', '3px', 'important')
  contrast.style.setProperty('height', '3px', 'important')
  contrast.style.setProperty('margin', '3px 0', 'important')
  contrast.style.setProperty('background', 'rgb(255,255,255)', 'important')

  const transform = {value: 'none', important: true}
  const html = __setElementStyleProperties([
    document.documentElement,
    {transform, '-webkit-transform': transform},
  ])
  const body = __setElementStyleProperties([
    document.body,
    {transform, '-webkit-transform': transform},
  ])

  document.documentElement.setAttribute('data-applitools-original-transforms', JSON.stringify(html))
  document.body.setAttribute('data-applitools-original-transforms', JSON.stringify(body))
  return {
    offset: 1 * window.devicePixelRatio,
    size: 3 * window.devicePixelRatio,
    mask: [0, 1, 0],
  }
}
""",
)

_CLEANUP_PAGE_MARKER_JS = "{} {} return __cleanupPageMarker()".format(
    _SET_ELEMENT_STYLE_PROPERTIES_JS,
    """
function __cleanupPageMarker() {
  const marker = document.querySelector('[data-applitools-marker]')
  if (marker) document.body.removeChild(marker)
  const html = document.documentElement.getAttribute('data-applitools-original-transforms')
  const body = document.body.getAttribute('data-applitools-original-transforms')
  if (html) __setElementStyleProperties([document.documentElement, JSON.parse(html)])
  if (body) __setElementStyleProperties([document.body, JSON.parse(body)])
}
""",
)



from collections import Counter
from typing import TYPE_CHECKING, Union

from applitools.common import (
    DiffsFoundError,
    NewTestError,
    TestFailedError,
    TestResultContainer,
    TestResults,
    logger,
)
from applitools.common.utils import datetime_utils, iteritems

if TYPE_CHECKING:
    from typing import Callable, Dict, List

    from applitools.selenium.visual_grid import RunningTest


def wait_till_tests_completed(test_provider):
    # type: (Union[Callable, List]) -> None
    def get_tests(provider):
        if isinstance(test_provider, list):
            return test_provider
        return test_provider()

    while True:
        states = list(set(t.state for t in get_tests(test_provider)))
        if not states:
            # probably some exception is happened during execution
            break
        if len(states) == 1 and states[0] == "completed":
            break
        datetime_utils.sleep(1500, msg="Waiting for state completed!")


def collect_test_results(tests, should_raise_exception):
    # type: (Dict[RunningTest, TestResults], bool) -> List[TestResultContainer]
    all_results = []
    for test, test_result in iteritems(tests):
        if test.pending_exceptions:
            logger.error(
                "During test execution above exception raised. \n {:s}".join(
                    str(e) for e in test.pending_exceptions
                )
            )
        if test.has_checks:
            exception = None
        else:
            exception = TestFailedError("Test has no checks")
        if test_result:
            scenario_id_or_name = test_result.name
            app_id_or_name = test_result.app_name
            if test_result.is_unresolved and not test_result.is_new:
                exception = DiffsFoundError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result.is_new:
                exception = NewTestError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
            if test_result.is_failed:
                exception = TestFailedError(
                    test_result, scenario_id_or_name, app_id_or_name
                )
        else:
            exception = TestFailedError("Test haven't finished correctly")
        all_results.append(
            TestResultContainer(test_result, test.browser_info, exception)
        )
        if exception and should_raise_exception:
            raise exception
    return all_results



from enum import Enum
from typing import Text

import attr
from ua_parser import user_agent_parser

__all__ = ("parse_user_agent_string", "BrowserNames", "OSNames", "UserAgent")


def parse_user_agent_string(ua_string):
    # type: (str) -> UserAgent
    parsed = user_agent_parser.Parse(ua_string)
    ua = parsed["user_agent"]
    os = parsed["os"]
    try:
        browser_name = BrowserNames(ua["family"])
    except ValueError:
        browser_name = BrowserNames.Unknown
    try:
        os_name = OSNames(os["family"])
    except ValueError:
        os_name = OSNames.Unknown
    return UserAgent(
        os=os_name,
        os_major_version=os["major"],
        os_minor_version=os["minor"],
        browser=browser_name,
        browser_major_version=ua["major"],
        browser_minor_version=ua["minor"],
        origin_ua_string=ua_string,
    )


def to_float(val):
    try:
        return float(val)
    except TypeError:
        return -1


class BrowserNames(Enum):
    Unknown = "Unknown"
    Edge = "Edge"
    IE = "IE"
    Firefox = "Firefox"
    Chrome = "Chrome"
    ChromeMobile = "Chrome Mobile"
    HeadlessChrome = "HeadlessChrome"
    Safari = "Safari"
    Chromium = "Chromium"
    MobileSafari = "Mobile Safari"


class OSNames(Enum):
    Unknown = "Unknown"
    Windows = "Windows"
    IOS = "iOS"
    Macintosh = "Macintosh"
    ChromeOS = "ChromeOS"
    Android = "Android"
    Linux = "Linux"


@attr.s
class UserAgent(object):
    os = attr.ib(default=OSNames.Unknown, converter=OSNames)
    os_major_version = attr.ib(default=-1, converter=to_float)
    os_minor_version = attr.ib(default=-1, converter=to_float)
    browser = attr.ib(default=BrowserNames.Unknown, converter=BrowserNames)
    browser_major_version = attr.ib(default=-1, converter=to_float)
    browser_minor_version = attr.ib(default=-1, converter=to_float)
    origin_ua_string = attr.ib(default=None)

    @property
    def is_internet_explorer(self):
        # type: () -> bool
        return bool(
            self.browser == BrowserNames.IE
            or (self.browser == BrowserNames.Edge and self.browser_major_version <= 18)
        )

    @property
    def is_chrome(self):
        # type: () -> bool
        return self.browser in [
            BrowserNames.Chrome,
            BrowserNames.Chromium,
            BrowserNames.ChromeMobile,
        ]

    @property
    def os_name_with_major_version(self):
        # type: () -> Text
        if self.os_major_version != -1:
            return "{} {:.0f}".format(self.os.value, self.os_major_version)
        return self.os.value
